import { Component, Input, TemplateRef, Optional, OnInit } from '@angular/core';
import { FormGroup, FormGroupName, AbstractControl, ControlContainer } from '@angular/forms';

export interface DynarowContext {
	control: AbstractControl;
	meta: StringMap;
}

@Component({
	selector: 'app-dynaform',
	templateUrl: './dynaform.component.html',
	styleUrls: ['./dynaform.component.scss']
})
export class DynaformComponent implements OnInit {

	/*
	 * DynaformComponent: <app-dynaform>
	 *
	 * USAGE:
	 *
	 * Supply component with either a FormGroup or the name of a FormGroup,
	 * the forms full metadata tree,
	 * and, optionally, a TemplateRef used as a field row template (overrides the default template)
	 *
	 * e.g.
	 * <app-dynaform [formGroup]="myForm" [meta]="myFormMetaData"></app-dynaform>
	 * <app-dynaform formGroupName="myFormGroupName" [meta]="myFormMetaData"></app-dynaform>
	 * <app-dynaform [formGroup]="myForm" [meta]="myFormMetaData" [template]="myFieldTemplate"></app-dynaform>
	 *
	 * If supplied with just a FormGroupName it will retieve the FormGroup from the injected ControlContainer
	 *
	 */

	@Input()
	formGroup?: FormGroup;

	@Input()
	formGroupName?: FormGroupName;

	@Input()
	set control(fg: FormGroup) {
		this.formGroup = fg;
	}

	@Input()
	set meta(data) {
		this.formMetaData = data;
	}

	@Input()
	template?: TemplateRef<any>;

	formMetaData: any; // TODO: Tighten up type
	controlNames: string[];

	constructor(
		@Optional() private cc: ControlContainer
	) {}

	ngOnInit() {
		
		// Get the full path in the tree (used to match metadata)
		let parents = [];
		if (this.formGroup) {
			// formGroup supplied directly, so reconstruct current FormGroup's full path
			let fg = this.formGroup;
			while (fg.parent) {
				// Find the identity (i.e. object key) of the parent FormGroup that IS FormGroup 'fg'
				const fgIdentity = Object.entries(fg.parent.controls).filter(([key, candidate]) => candidate === fg);
				parents.push(fgIdentity[0][0]);
				fg = <FormGroup>fg.parent;
			}
		} else if (this.formGroupName) {
			// formGroupName supplied, so get the full path from the ControlContainer
			parents = this.cc.path.reverse();
		}
		const path = parents;
		
		// Get the formGroup from the formGroupName if necessary
		if (!this.formGroup && this.formGroupName) {
			this.formGroup = <FormGroup>this.cc.control; // Get theFormGroup from the ControlContainer
		}
		if (!this.formGroup) {
			throw new Error('Dynaform Component initialised without [formGroup] or formGroupName');
		}
		this.controlNames = Object.keys(this.formGroup.controls);

		// If we're given a formGroupName or nested FormFroup, and the form's full (or partial but fuller) metadata tree,
		// drill down to find the corresponding FormGroup's metadata
		const metaDataKeysExpected = this.controlNames.join(',');
		while (path.length && metaDataKeysExpected !== Object.keys(this.formMetaData).join(',')) {
			const branch = path.pop();
			this.formMetaData = this.formMetaData[branch].meta;
		}
		
		// Check we've got a match
		const metaDataKeys = Object.keys(this.formMetaData).join(',');
		if (metaDataKeys !== metaDataKeysExpected) {
			throw new Error(`
				Dynaform can't match FormGroup's controls with metadata
				Expected ${metaDataKeysExpected}
				Got ${metaDataKeys}`
			);
		}
	}

	getTemplateContext(controlName: string): DynarowContext {
		return {
			control: this.formGroup.get(controlName),
			meta: this.formMetaData[controlName]
		};
	}

	getRowClass(inputType: string): string {
		return `row-${inputType.toLowerCase().replace('component', '')}`;
	}

}
